import os
import time
import sys
import argparse
from pathlib import Path

from ImmunoViewer.helpers.process_ome_tiff import process_ome_tiff
from ImmunoViewer.helpers.process_tiff import process_tiff

def main():
    print("Starting ImmunoViewer watch folder")
    parser = argparse.ArgumentParser(description="Watch folder and process new files as they are added")
    parser.add_argument('import_dir', nargs='?', help='The data directory to watch for new files (default: %(default)s)', default='/iv-import')
    parser.add_argument('storage_dir', nargs='?', help='The output directory for the processed files (default: %(default)s)', default='/iv-store')
    parser.add_argument('vips', nargs='?', help='Path to the VIPS executable (default: %(default)s)', default='vips')
    args = parser.parse_args()

    if not args.import_dir or not args.storage_dir:
        print("Usage: process_folder.py [-t num_cores] import_dir storage_dir vips")
        sys.exit(1)
    
    import_dir = Path(args.import_dir)
    storage_dir = Path(args.storage_dir)
    vips = args.vips

    print(f"Watching folder {import_dir} for new files, output to {storage_dir}, and vips location is {vips}")

    while True: 
        # iterate .ome.tiff files in import_dir
        for file in import_dir.glob('*.ome.tiff'):
            # if no folder with same name as ome.tiff file (without .tiff) is found in storage_dir, run export
            if not os.path.exists(storage_dir / file.stem):
                # if .ome.tiff file has not been modified for 10 seconds, run export
                if time.time() - os.path.getmtime(file) > 10:
                    process_ome_tiff(file, storage_dir, vips)
        
        # iterate all folders in import_dir
        for folder in import_dir.iterdir():
            # iterate .tiff or .tif files in folder
            for file in folder.glob('*.tif*'):
                # if no folder with same name as tiff file (without .tiff) is found in storage_dir, run export
                relative_folder_path = folder.relative_to(import_dir)
                target_dir = storage_dir / relative_folder_path / f'{file.stem}_files'
                # print("checking for: ", target_dir)
                if not os.path.exists(target_dir):
                    # if tiff file has not been modified for 10 seconds, run export
                    if time.time() - os.path.getmtime(file) > 60:
                        process_tiff(file, storage_dir, vips)
                        # print("processing..", file, storage_dir, vips)
                
        time.sleep(30)

if __name__ == "__main__":
    main()



from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("epi_models")
except PackageNotFoundError:
    # If the package is not installed, don't add __version__
    pass


import os
import time
import sys
import argparse
from pathlib import Path
from threading import Thread

from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

from helpers.process_ome_tiff import process_ome_tiff
from helpers.process_tiff import process_tiff

class Handler(FileSystemEventHandler):
    def __init__(self):
        super().__init__()
        self.last_modified = {}

    def on_any_event(self, event):
        if event.is_directory:
            return None
        elif event.event_type == 'created':
            print("Watchdog received created event - % s." % event.src_path)
        elif event.event_type == 'modified':
            print("Watchdog received modified event - % s." % event.src_path)

        self.last_modified[event.src_path] = time.time()

def check_files(handler, import_dir, storage_dir, vips):
    '''
    Check the last modified time of files and process them if they haven't been modified for 10 seconds.
    This is done to ensure the file is done uploading before processing.
    '''
    while True:
        current_time = time.time()
        for file_path, last_modified_time in handler.last_modified.items():
            if current_time - last_modified_time > 10:
                print(f"File {file_path} hasn't been modified for 10 seconds.")
                # process ome-tiff files
                if file_path.endswith('.ome.tiff') or file_path.endswith('.ome.tif'):
                    process_ome_tiff(file_path, storage_dir, vips)
                  
                # process tiff files

                if file_path.endswith('.tiff') or file_path.endswith('.tif'):
                  relative_path = os.path.relpath(file_path, import_dir)

                  if os.path.sep in relative_path:
                    print(f"File {file_path} is in a subfolder.")
                    # process the file in the subfolder
                    process_tiff(file_path, storage_dir, vips)

                handler.last_modified.pop(file_path)
                break
        time.sleep(10)

def main():
    print("Starting ImmunoViewer watch folder")
    parser = argparse.ArgumentParser(description="Watch folder and process new files as they are added")
    parser.add_argument('import_dir', nargs='?', help='The data directory to watch for new files (default: %(default)s)', default='/iv-import')
    parser.add_argument('storage_dir', nargs='?', help='The output directory for the processed files (default: %(default)s)', default='/iv-store')
    parser.add_argument('vips', nargs='?', help='Path to the VIPS executable (default: %(default)s)', default='vips')
    parser.add_argument('-t', '--num_cores', type=int, default=4, help='Number of cores to use (default: %(default)s)')
    args = parser.parse_args()

    if not args.import_dir or not args.storage_dir:
        print("Usage: process_folder.py [-t num_cores] import_dir storage_dir vips")
        sys.exit(1)
    
    num_cores = args.num_cores if args.num_cores else 4
    import_dir = Path(args.import_dir)
    storage_dir = Path(args.storage_dir)
    vips = args.vips

    print(f"Watching folder {import_dir} for new files, output to {storage_dir}, and vips location is {vips}")

    handler = Handler()

    observer = Observer()
    observer.schedule(handler, import_dir, recursive=True)
    observer.start()

    # Start the thread for periodic file checking
    file_check_thread = Thread(target=check_files, args=(handler, import_dir, storage_dir, vips))
    file_check_thread.daemon = True
    file_check_thread.start()

    try:
        while True:
            pass
    except KeyboardInterrupt:
        observer.stop()

    observer.join()

if __name__ == "__main__":
    main()



import pathlib
from io import BytesIO
import os
import json
import cv2
import numpy as np
from pydantic_settings import BaseSettings

from ImmunoViewer.helpers.server_helpers import *

from fastapi import FastAPI, Request, HTTPException, Path, status
from fastapi.responses import FileResponse, JSONResponse, Response, PlainTextResponse
from fastapi.staticfiles import StaticFiles
from fastapi.middleware.cors import CORSMiddleware

import uvicorn
import argparse

app = FastAPI()

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

class Settings(BaseSettings):
    SLIDE_DIR: str = "/iv-store"
    SAVE: bool = False
    COLORS: dict = {
        'green': (0, 255, 0),
        'red': (0, 0, 255),
        'blue': (255, 0, 0),
        'yellow': (0, 255, 255),
        'cyan': (255, 255, 0),
        'magenta': (255, 0, 255),
        'white': (255, 255, 255),
        'black': (0, 0, 0),
    }
    class Config:
        env_prefix = "IV_"

settings = Settings()

def main(host="127.0.0.1", port=8000, reload=False):
    """Run the API server with Uvicorn."""
    uvicorn.run("ImmunoViewer.server:app", host=host, port=port, reload=reload)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run ImmunoViewer server.")
    parser.add_argument("--host", type=str, default="127.0.0.1", help="Host for the API server")
    parser.add_argument("--port", type=int, default=8000, help="Port for the API server")
    parser.add_argument("--reload", action="store_true", help="Enable automatic reload")
    parser.add_argument("--save", action="store_true", help="Enable saving of slide settings")
    parser.add_argument("--slide-dir", type=str, help="Directory to store slide files")
    args = parser.parse_args()

    if args.save:
        settings.SAVE = True

    if args.slide_dir:
        settings.SLIDE_DIR = args.slide_dir
        
    main(host=args.host, port=args.port, reload=args.reload)

current_folder = pathlib.Path(__file__).parent.resolve()
client_dir = os.path.join(current_folder, "client")
slide_dir = pathlib.Path(settings.SLIDE_DIR)

app.mount("/assets", StaticFiles(directory=os.path.join(client_dir, "assets")), name="assets")
app.mount("/images", StaticFiles(directory=os.path.join(client_dir, "images")), name="images")

@app.get("/favicon.ico", include_in_schema=False)
async def favicon():
    return FileResponse(os.path.join(client_dir, 'favicon.ico'))

@app.get("/", include_in_schema=False)
async def read_root():
    # Serve index.html for the root
    return FileResponse(os.path.join(client_dir, 'index.html'))

@app.get('/samples.json')
async def samples():
    print("looking in ", os.path.abspath(settings.SLIDE_DIR))
    try:
        file_json = find_directories_with_files_folders(os.path.abspath(settings.SLIDE_DIR))
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

    buf = {
        "samples": file_json,
        "save": settings.SAVE,
        "colors": list(settings.COLORS.keys())
    }

    return JSONResponse(content=buf, status_code=200)

@app.get("/{files}/{chs}/{gains}/{file}.dzi")
async def get_dzi(files: str, chs: str, gains: str, file: str):
    """
    Get the DZI file
    """
    files = files.split(';')
    path_to_dzi = os.path.join(os.path.abspath(settings.SLIDE_DIR), file, f"{files[0].replace('_files', '')}.dzi")
    
    try:
        dzi_content = modify_dzi_format(path_to_dzi)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

    return Response(content=dzi_content, media_type="application/xml", status_code=200)

@app.get("/{files}/{chs}/{gains}/{file}_files/{level}/{loc}")
async def get_tile(
    files: str, 
    chs: str, 
    gains: str, 
    file: str = Path(..., description="The base name of the file without extension"),
    level: int = Path(..., description="The image pyramid level"), 
    loc: str = Path(..., description="Location of the tile with extension")
):
    """
    Get a tile from the slide
    """

    loc = loc.replace('.jpeg', '.tiff')
    files = files.split(';') 
    chs = chs.split(';')
    gains = gains.split(';')
    gains = [int(x) for x in gains]

    if len(files) == 1:
        file_path = slide_dir / file / f'{files[0]}' / str(level) / loc
        merged_image = cv2.imread(str(file_path))
    else:
        merge_images = []
        merge_chs = []
        merge_gains = []



        for i, ch in enumerate(chs):
            if ch != 'empty':

                file_path = slide_dir / file / f'{files[i]}' / str(level) / loc
                buf = cv2.imread(str(file_path))

                buf = cv2.cvtColor(buf, cv2.COLOR_BGR2GRAY)
                merge_images.append(buf)
                merge_chs.append(ch)
                merge_gains.append(gains[i])


        if len(merge_images) == 0:
            merged_image = np.zeros((256, 256), dtype=np.uint8)
        else:
            merged_image = mix_channels(merge_images, merge_chs, merge_gains, settings.COLORS)

    img_bytes = cv2.imencode('.jpeg', merged_image)[1].tobytes()
    img_io = BytesIO(img_bytes)

    return Response(content=img_io.getvalue(), media_type='image/jpeg')

@app.post("/save/{file:path}", response_class=PlainTextResponse)
async def save_slide_settings(file: str, request: Request):
    """
    Save the slide settings
    """
    if settings.SAVE:
        try:
            data = await request.json()
        except json.JSONDecodeError:
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Invalid JSON")

        file_path = slide_dir / file / 'sample.json'

        with open(file_path, 'w') as f:
            json.dump(data, f)

        return "OK"
    else:
        return PlainTextResponse("SAVE BLOCKED", status_code=status.HTTP_403_FORBIDDEN)


#!/usr/bin/env python3

from aicsimageio import AICSImage
import tifffile
import os
# import pyvips
import json

def process_tiff(input_tiff_file, output_folder, vips):
    """
    Process a folder with tiff files, saving each channel as an individual TIFF file
    and converting them to Deep Zoom Image (DZI) format.
    
    Args:
    - ome_tiff_path (str): Path to the folder.
    - base_folder (str): Base folder to save the output files.
    """
    
    base_folder = os.path.join(output_folder, os.path.basename(os.path.dirname(input_tiff_file)))
    file_name = os.path.splitext(os.path.basename(input_tiff_file))[0]

    print(f"Processing file {input_tiff_file} to {base_folder}")

    # see if directory exists
    if not os.path.exists(base_folder):
        os.makedirs(base_folder)

    command = f'{vips} dzsave {input_tiff_file} {base_folder}/{file_name} --suffix .tiff'
    print(command)
    os.system(command)

    # Dump json with slide info
    print(f"Creating JSON file for slide info")
    # try open json 
    try:
        with open(f'{base_folder}/sample.json', 'r') as f:
            data = json.load(f)
    except:
        data = {
            "ch": {},
            "gain": {},
            "ch_stain": {},
            "description": "",
            "overlays": []
        }
    
    # add new data
    colors = ["blue", "green", "green", "yellow", "cyan"]
    filler_gain_value = 6

    length = len(data["ch"])
    if length < len(colors):
        data["ch"][file_name + "_files"] = colors[length]
    else:
        data["ch"][file_name + "_files"] = "empty"
    
    data["gain"][file_name + "_files"] = filler_gain_value
    data["ch_stain"][file_name + "_files"] = file_name

    with open(f'{base_folder}/sample.json', 'w') as f:
        json.dump(data, f)


if __name__ == "__main__":
    # Example usage
    input_tiff_file = '/Volumes/DATA/iv-import/Q129_S005_A107_CosMxTMA2__2024-03-14-06-18_EM001_000491.ome.tiff'
    output_folder = "/Volumes/DATA/iv-store"
    vips = '/opt/homebrew/bin/vips'
    process_tiff(input_tiff_file, output_folder, vips)


#!/usr/bin/env python3

from aicsimageio import AICSImage
import tifffile
import os
# import pyvips
import json

def process_ome_tiff(ome_tiff_path, output_folder, vips):
    """
    Process an OME-TIFF file, saving each channel as an individual TIFF file
    and converting them to Deep Zoom Image (DZI) format.
    
    Args:
    - ome_tiff_path (str): Path to the OME-TIFF file.
    - base_folder (str): Base folder to save the output files.
    """
    
    base_folder = os.path.join(output_folder, os.path.splitext(os.path.basename(ome_tiff_path))[0])

    image = AICSImage(ome_tiff_path)
    n_channels = image.shape[1]

    # see if directory exists
    if not os.path.exists(base_folder):
        os.makedirs(base_folder)

    # Iterate over each channel and save it as an individual TIFF file
    print(f"Processing {n_channels} channels")

    for channel in range(n_channels):
        print(f"Processing channel {channel}")
        # Extract the channel data
        channel_data = image.get_image_data("YX", C=channel)
        
        # Save the channel data as a TIFF file
        output_file_name = f'{base_folder}/channel_{channel}.tiff'
        tifffile.imwrite(output_file_name, channel_data)

    # Inside the loop where you convert TIFF files to DZI
    print(f"Converting {n_channels} channels to DZI")
    for channel in range(n_channels):
        print(f"Converting channel {channel} to DZI")
        command = f'{vips} dzsave {base_folder}/channel_{channel}.tiff {base_folder}/channel_{channel} --suffix .tiff'
        print(command)
        os.system(command)

    # Dump json with slide info 
    print(f"Creating JSON file for slide info")
    channel_details = image.metadata.images[0].pixels.channels
    colors = ["blue", "green", "green", "yellow", "cyan"]
    image_meta = {
        "ch": {},
        "gain": {},
        "ch_stain": {},
        "description": "",
        "overlays": []
    }
    filler_gain_value = 6

    for i, channel in enumerate(channel_details):
        if i < len(colors):
            image_meta["ch"]["channel_" + str(i) + "_files"] = colors[i]
        else:
            image_meta["ch"]["channel_" + str(i) + "_files"] = "empty"

        image_meta["gain"]["channel_" + str(i) + "_files"] = filler_gain_value
        image_meta["ch_stain"]["channel_" + str(i) + "_files"] = channel.name

    image_meta_json = json.dumps(image_meta)

    with open(base_folder + "/sample.json", 'w') as file:
        file.write(image_meta_json)


if __name__ == "__main__":
    # Example usage
    ome_tiff_path = '/Volumes/DATA/iv-import/Q129_S005_A107_CosMxTMA2__2024-03-14-06-18_EM001_000491.ome.tiff'
    output_folder = "/Volumes/DATA/iv-store"
    vips = '/opt/homebrew/bin/vips'
    process_ome_tiff(ome_tiff_path, output_folder, vips)


import os
import json
import cv2
import numpy as np
from xml.etree import ElementTree as ET
from concurrent.futures import ThreadPoolExecutor
import asyncio

def mix_channels(images, chs, gains, color_mapping) -> np.ndarray:
    """
    Mixes multiple images into one by applying color mapping to each channel
    :param images: list of images to mix
    :param chs: list of channels to mix
    :param gains: list of gains to apply to each channel
    :param color_mapping: dictionary with color mapping
    :return: merged image
    """

    merged_image = np.zeros((*images[0].shape, 3), dtype=np.uint8)

    for image, ch, gain in zip(images, chs, gains):
        color_rgb = color_mapping[ch]
        enhanced_image = cv2.convertScaleAbs(image, alpha=gain)
        colored_image = cv2.merge([enhanced_image * (color_rgb[i] // 255) for i in range(3)])
        merged_image = cv2.add(merged_image, colored_image)

    return merged_image

def find_directories_with_files_folders(base_dir) -> list:
    """
    Finds all directories with _files suffix and returns them as a list of dictionaries
    :param base_dir: base directory to search in
    :return: list of dictionaries with directory name and files
    """
    directories_with_files = []

    for root, dirs, files in os.walk(base_dir):
        if root.count(os.sep) - base_dir.count(os.sep) >= 2:
            del dirs[:]
        for dirname in dirs:
            if dirname.endswith("_files"):
                buf = {}
                buf['name'] = os.path.relpath(root, base_dir)
                dirs.sort()
                buf['files'] = dirs
                directories_with_files.append(buf)

                buf['details'] = {}

                if os.path.isfile(os.path.join(f'{root}','sample.json')):
                        with open(os.path.join(f'{root}','sample.json'), 'r') as f:
                            data = json.load(f)
                            buf['details'] = data
                break
    return directories_with_files

def modify_dzi_format(dzi_file_path) -> str:
    """
    Modifies the DZI file format from TIFF to JPEG
    :param dzi_file_path: path to the DZI file
    :return: modified DZI file content
    """
    with open(dzi_file_path, 'r') as file:
        dzi_content = file.read()

    tree = ET.ElementTree(ET.fromstring(dzi_content))
    root = tree.getroot()

    if root.attrib['Format'].lower() == 'tiff':
        root.attrib['Format'] = 'jpeg'

    modified_dzi_str = ET.tostring(root, encoding='unicode')
    
    return modified_dzi_str